package FI2.Stream;
/**{p}
 * -- FI2.Stream --
 * Universita' di Padova -
 * Diploma Universitario in Ingegneria Informatica
 * Corso di Fondamenti di Informatica 2
 * Flussi
 */

/**{f}
 * ReduceSpace.java
 * esempio di utilizzo di stream come filtro di 
 * eliminazione 
 *
 * @version 1.00 98/10/02
 * @author M.Moro DEI UPD
 */

import java.io.*;
import FI2.Util.FI2Sys;

/**{c}
 * legge da uno stream di input singoli caratteri e, per ogni
 * riga, elimina eventuali spazi iniziali e finali e riduce
 * a uno gli spazi di separazione delle parole -
 * le routine di lettura/scrittura consentono poi di
 * manipolare char anziche' int -
 */
public class ReduceSpace
{
    private static final char EOFCHAR = 0xffff;
      // corrisponde al -1 ritornato da read()
      // non corrisponde ad alcun carattere ASCII valido
    private InputStream in;  // stream di ingresso
    private OutputStream out;  // stream di uscita

    /**[c]
     * costruttore di default -
     * effettua tutta l'operazione di filtro
     * riduzione tra i due stream
     * @param inS  stream di input
     * @param outS  stream di output
     */
    public ReduceSpace(InputStream inS, OutputStream outS)
    {
        in = inS;
        out = outS;
        try
        {
            while (reduceLine() != EOFCHAR);
        }
        catch (IOException e)
        {
            // nulla da fare
        }
    } //[c] ReduceSpace

    /**[m]
     * verifica se carattere di spaziatura
     * @param c  carattere
     * @return true se di spaziatura, false altrimenti
     */
    private boolean isSpace(char c)
    { return Character.isSpaceChar(c); }

    /**[m]
     * acquisisce un carattere dallo stream di input
     * @return carattere acquisito
     */
    private char getChar()
      throws IOException
    { return (char)in.read(); }

    /**[m]
     * restituisce un carattere allo stream di input
     * @param c  carattere da restituire
     */
    private void ungetChar(char c)
      throws IOException
    {
        if (!(in instanceof PushbackInputStream))
          // non e` un pushback stream ->
          // se ne 'monta' uno sopra
            in = new PushbackInputStream(in);
        ((PushbackInputStream)in).unread((int)c);
    } //[m] ungetChar

    /**[m]
     * emette un carattere sullo stream di output
     * @param c  carattere da emettere
     */
    private void putChar(char c)
      throws IOException
    { out.write((int)c); }

    /**[m]
     * operazioni di fine linea
     * @param c  carattere terminatore
     */
    private void endLine(char c)
      throws IOException
    {
        if (c == '\n')
        {
            // la riga e` finita
            putChar(c);
            return;
        } //if \n
        else if (c == '\r')
        {
            // la riga e` finita
            putChar(c);
            if ((c=getChar()) != '\n')
                // carriage ret. non seguito da newline
                ungetChar(c);
            else
                putChar(c);
            return;
        } //if \r
    } //[m] endLine

    /**[m]
     * riduzione degli spazi per una linea
     * @return il carattere terminatore
     */
/*
Questo metodo simula il DFA di riconoscimento i cui archi
sono etichettati con il carattere in lettura e con l'azione
da svolgere (secondo il modello di Mealy). Il DFA e` cosi`
descritto:

Stati:
S1  iniziale, spazi iniziali
W   all'inetrno di una parola
S2  spazi di separazione o finali
E   stato finale, fine riga

Archi:

Stato     Stato     Carattere    Azione
partenza  arrivo    input (car)
--------------------------------------
S1        S1        space        --
          W         normale      <car>
          E         newline      \n  (con elim. eventuale \r)
          E         EOF          --
W         W         normale      <car>
          S2        space        --
          E         newline      \n   (con elim. eventuale \r)
          E         EOF          --
S2        S2        space        --
          W         normale      spazio <car>
          E         newline      \n   (con elim. eventuale \r)
          E         EOF          --
*/
    private char reduceLine()
      throws IOException
    {
        char c;

        // stato S1
        while (isSpace(c = getChar()) && c != EOFCHAR);
          // elimina spazi iniziali
        if (c == '\n' || c == '\r' || c == EOFCHAR)
        {
            // la riga e` finita
            endLine(c);
            return c;
        }
        // carattere normale
        putChar(c);
        
        // stato W
        for (; ; )
        {
            while (!isSpace(c=getChar()) && 
              c != '\r' && c != '\n'  && c != EOFCHAR)
                // carattere normale
                putChar(c);
            if (!isSpace(c))
              // fine riga
            {
                endLine(c);
                return c;
            }

            // stato S2
            while (isSpace(c=getChar()) && c != EOFCHAR);
              // salta spazi
            if (c=='\r' || c=='\n' || c == EOFCHAR)
              // fine riga o fine file
            {
                endLine(c);
                return c;
            }
            // un solo spazio emesso
            putChar(' ');
            putChar(c);
        } //for
    } //[m] reduceLine


    /**[m][s]
     * metodo di collaudo
     * @param args[0]  opzionale: nome file di input
     * @param args[1]  opzionale: nome file di output
     */
    public static void main(String[] args)
      throws IOException
    {
        switch (args.length)
        {
        case 0:
            System.err.println(
              "riduzione spazi da stdin a stdout");
            new ReduceSpace(System.in, System.out);
            System.out.flush();  // garantisce l'output
            break;
        case 1:
            // solo input
            System.err.println(
              "riduzione spazi da " + args[0] +
              " a stdout");
            InputStream in = new BufferedInputStream(
              new FileInputStream(args[0]));
            new ReduceSpace(in, System.out);
            in.close();
            break;
        default :
            // solo input e output
            System.err.println(
              "riduzione spazi da " + args[0] +
              " a " + args[1]);
            in = new BufferedInputStream(
              new FileInputStream(args[0]));
            OutputStream out = new BufferedOutputStream(
              new FileOutputStream(args[1]));
            new ReduceSpace(in, out);
            in.close();
            out.close();
        } //switch
    } //[m][s] main

} //{c} ReduceSpace
